
let pi = 4. *. atan 1.

(* 複素数の計算 *)
let ( +@ ) z1 z2 = Complex.add z1 z2
let ( -@ ) z1 z2 = Complex.sub z1 z2
let ( *@ ) z1 z2 = Complex.mul z1 z2

(* ビット反転アルゴリズム *)
let mod_2p n p = (n asr p) mod 2
let bit_reversal_index n p =
  Array.init p (fun k -> mod_2p n k) |> Array.to_list |> List.rev |> List.mapi (fun k b -> if b = 1 then (1 lsl k) else 0) |> List.fold_left (+) 0

let bit_reversal p x = Array.init (Array.length x) (fun k -> x.(bit_reversal_index k p)) 

(* 回転子 *)
let w p n =
  let t = -2. *. pi *. (float n) /. (float (1 lsl p)) in
  { Complex.re = cos t; im = sin t }

let wp p n =
  let t = 2. *. pi *. (float n) /. (float (1 lsl p)) in
  { Complex.re = cos t; im = sin t }
    
    
let butterfly n m w x =
  let xn = x.(n) in
  let xm = x.(m) in

  x.(n) <- xn +@ (w *@ xm);
  x.(m) <- xn -@ (w *@ xm);
  ()


let e4 k = Array.init 4 (fun n -> wp 2 (n*k))
let e8 k = Array.init 8 (fun n -> wp 3 (n*k))
let e16 k = Array.init 16 (fun n -> wp 4 (n*k))
let e p k = Array.init (1 lsl p) (fun n -> wp p (n*k)) 

(* 4 点 FFT *)
let fft4 x =
  let y = bit_reversal 2 x in
  (* 1st *)
  butterfly 0 1 (w 1 0) y;
  butterfly 2 3 (w 1 0) y;
  (* 2nd *)
  butterfly 0 2 (w 2 0) y;
  butterfly 1 3 (w 2 1) y;
  
  y

let fft4' x =
  let y = bit_reversal 2 x in
  (* 1st *)
  for n = 0 to (1 lsl 1) - 1 do
    butterfly (n*2) (n*2+1) (w 1 0) y;
  done;
  (* 2nd *)
  for r = 0 to (1 lsl 1) - 1 do
    butterfly r 2 (w 2 r) y;
  done;
  
  y


(* 8 点 FFT *)
let fft8 x =
  let y = bit_reversal 3 x in 
  
  (* 1st *)
  butterfly 0 1 (w 1 0) y;
  butterfly 2 3 (w 1 0) y;
  butterfly 4 5 (w 1 0) y;
  butterfly 6 7 (w 1 0) y;

  (* 2nd *)
  butterfly 0 2 (w 2 0) y;
  butterfly 1 3 (w 2 1) y;
  butterfly 4 6 (w 2 0) y;
  butterfly 5 7 (w 2 1) y;

  (* 3rd *)
  butterfly 0 4 (w 3 0) y;
  butterfly 1 5 (w 3 1) y;
  butterfly 2 6 (w 3 2) y;
  butterfly 3 7 (w 3 3) y;

  y

(* for 文で簡略化した 8 点 FFT *)
let fft8' x =
  let y = bit_reversal 3 x in 
  
  (* 1st *)
  for n = 0 to (1 lsl 2)-1 do 
    butterfly (2*n) (2*n+1) (w 1 0) y;
  done;

  (* 2nd *)
  for n = 0 to (1 lsl 1)-1 do
    for r = 0 to (1 lsl 1)-1 do
      butterfly (4*n+r) (4*n+2+r) (w 2 r) y;
    done;
  done;

  (* 3rd *)
  for r = 0 to (1 lsl 2) - 1 do
    butterfly r (4+r) (w 3 r) y;
  done;

  y


(* 16 点 FFT *)
let fft16 x =
  let y = bit_reversal 4 x in

  (* 1st *)
  butterfly 0 1 (w 1 0) y;
  butterfly 2 3 (w 1 0) y;
  butterfly 4 5 (w 1 0) y;
  butterfly 6 7 (w 1 0) y;
  butterfly 8 9 (w 1 0) y;
  butterfly 10 11 (w 1 0) y;
  butterfly 12 13 (w 1 0) y;
  butterfly 14 15 (w 1 0) y;

  (* 2nd *)
  butterfly 0 2 (w 2 0) y;
  butterfly 1 3 (w 2 1) y;
  butterfly 4 6 (w 2 0) y;
  butterfly 5 7 (w 2 1) y;
  butterfly 8 10 (w 2 0) y;
  butterfly 9 11 (w 2 1) y;
  butterfly 12 14 (w 2 0) y;
  butterfly 13 15 (w 2 1) y;

  (* 3rd *)
  butterfly 0 4 (w 3 0) y;
  butterfly 1 5 (w 3 1) y;
  butterfly 2 6 (w 3 2) y;
  butterfly 3 7 (w 3 3) y;
  butterfly 8 12 (w 3 0) y;
  butterfly 9 13 (w 3 1) y;
  butterfly 10 14 (w 3 2) y;
  butterfly 11 15 (w 3 3) y;

  (* 4th *)
  butterfly 0 8 (w 4 0) y;
  butterfly 1 9 (w 4 1) y;
  butterfly 2 10 (w 4 2) y;
  butterfly 3 11 (w 4 3) y;
  butterfly 4 12 (w 4 4) y;
  butterfly 5 13 (w 4 5) y;
  butterfly 6 14 (w 4 6) y;
  butterfly 7 15 (w 4 7) y;

  y

(* for 文で簡略化した FFT *)
let fft16' x =
  let y = bit_reversal 4 x in

  (* 1st *)
  for n = 0 to (1 lsl 3) - 1 do
    for r = 0 to 0 do
      butterfly (2*n) (2*n+1+r) (w 1 0) y;
    done;
  done;

  (* 2nd *)
  for n = 0 to (1 lsl 2) - 1 do
    for r = 0 to (1 lsl 1) - 1 do
      butterfly (4*n+r) (4*n+2+r) (w 2 r) y;
    done;
  done;

  (* 3rd *)
  for n = 0 to (1 lsl 1) - 1 do
    for r = 0 to (1 lsl 2) - 1 do
      butterfly (8*n+r) (8*n+4+r) (w 3 r) y;
    done;
  done;

  (* 4th *)
  for n = 0 to 0 do
    for r = 0 to (1 lsl 3) - 1 do
      butterfly (16*n+r) (16*n+8+r) (w 4 r) y;
    done;
  done;

  y


(* 2^p 点の FFT *)
let fft m x =
  let y = bit_reversal m x in

  (* butterfly *)
  for p = 0 to m-1 do
    for n = 0 to (1 lsl (m-p-1)) - 1 do
      for r = 0 to (1 lsl p) - 1 do
	let ha = (1 lsl (p+1)) in
	let qu = (1 lsl p) in
	butterfly (ha*n+r) (ha*n+qu+r) (w (p+1) r) y;
      done;
    done;
  done;

  y


let fft' m x =
  let y = bit_reversal m x in
  let w_arr = Array.init (1 lsl m) (fun n -> w m n) in
  let w_val p n = w_arr.( (n lsl (m-p)) ) in
  
  for p = 0 to m-1 do
    let ns = (1 lsl (m-p-1)) - 1 in
    for n = 0 to ns do
      let rs = (1 lsl p) - 1 in
      for r = 0 to rs do
	let ha = (1 lsl (p+1)) in
	let qu = (1 lsl p) in
	(* print_endline ((string_of_int n) ^ " " ^ (string_of_int (ha*n+r)) ^ " " ^ (string_of_int (ha*n+qu+r))); *)
	butterfly (ha*n+r) (ha*n+qu+r) (w_val (p+1) r) y;
      done;
    done;
  done;

  y
    
