(*
 *
 *)

(* 
   3次元実射影空間
   [| x0; x1; x2; x3 |] (x3 <> 0)
*)

module Projective_3 : sig
  type t = private float array
  val of_array : float array -> t
  val to_array : t -> float array
  val motion_x : float -> t -> t
  val rot_x : float -> t -> t
  val motion_y : float -> t -> t
  val rot_y : float -> t -> t
  val motion_z : float -> t -> t
  val rot_z : float -> t -> t
end
= struct
  type t = float array
  let dim = 4
  let of_array v = assert (Array.length v = dim && v.(3) <> 0.); [| v.(0) /. v.(3); v.(1) /. v.(3); v.(2) /. v.(3); 1. |]
  let to_array v = [| v.(0) /. v.(3); v.(1) /. v.(3); v.(2) /. v.(3); 1. |]
  let inner_prod u v k =
    let uv = Array.init 4 (fun n -> u.(k).(n) *. v.(n)) in
    Array.fold_left (+.) 0. uv

  let prod m v = Array.init 4 (fun k -> inner_prod m v k)

  let a1 x = [|
    [|1.;0.;0.;x|];
    [|0.;1.;0.;0.|];
    [|0.;0.;1.;0.|];
    [|0.;0.;0.;1.|]
  |]

  let a2 y = [|
    [|1.;0.;0.;0.|];
    [|0.;1.;0.;y|];
    [|0.;0.;1.;0.|];
    [|0.;0.;0.;1.|]
  |]

  let a3 z = [|
    [|1.;0.;0.;0.|];
    [|0.;1.;0.;0.|];
    [|0.;0.;1.;z|];
    [|0.;0.;0.;1.|]
  |]

  let a4 phi = [|
    [|1.;0.;0.;0.|];
    [|0.;cos phi; -.sin phi;0.|];
    [|0.;sin phi; cos phi;  0.|];
    [|0.;0.;0.;1.|]
  |]

  let a5 psi = [|
    [|cos psi; 0.; sin psi; 0.|];
    [|0.; 1.; 0.; 0.|];
    [|-.sin psi;0.; cos psi;  0.|];
    [|0.;0.;0.;1.|]
  |]

  let a6 theta = [|
    [|cos theta; -.sin theta; 0.; 0.|];
    [|sin theta; cos theta; 0.; 0.|];
    [|0.;0.;1.;0.|];
    [|0.;0.;0.;1.|]
  |]


  let motion_x x p =
    let p_arr = to_array p in
    of_array (prod (a1 x) p_arr)
  
  let rot_x phi p =
    let p_arr = to_array p in
    of_array (prod (a4 phi) p_arr)

  let motion_y y p =
    let p_arr = to_array p in
    of_array (prod (a2 y) p_arr)
 
  let rot_y psi p =
    let p_arr = to_array p in
    of_array (prod (a5 psi) p_arr)

  let motion_z z p = 
    let p_arr = to_array p in
    of_array (prod (a3 z) p_arr)

  let rot_z theta p =
    let p_arr = to_array p in
    of_array (prod (a6 theta) p_arr)

end


(* *)
let init_gl width height =
  GlDraw.shade_model `smooth;
  GlClear.color (0.0, 0.0, 0.0);
  GlClear.depth 1.0;
  GlClear.clear [`color; `depth];
  Gl.enable `depth_test;
  GlFunc.depth_func `lequal;
  GlMisc.hint `perspective_correction `nicest

let z = 0.2
let pi = 4. *. atan 1.

let theta = ref 0.
let omega = 0.02 *. pi



let draw_gl_scene () =
  let dy = 1. +. cos !theta in
  let rec draw_cylinder z0 z1 r0 r1 th0 n idx =
    let s = 2. *. pi *. (float idx) /. (float n) in
    let (x0, y0) = (r0 *. cos (s +. th0), r0 *. sin (s +. th0) +. dy) in
    let (x1, y1) = (r1 *. cos (s +. th0), r1 *. sin (s +. th0) +. dy) in
    if idx >= n then
      begin
	GlDraw.vertex3(x0, y0, z0);
	GlDraw.vertex3(x1, y1, z1);
	GlDraw.ends()
      end
    else
      begin
	if idx = 0 then GlDraw.begins `quad_strip;
	GlDraw.vertex3(x0, y0, z0);
	GlDraw.vertex3(x1, y1, z1);
	draw_cylinder z0 z1 r0 r1 th0 n (idx+1)
      end
  in

  let rec draw_circle z r th0 n idx =
    let s = 2. *. pi *. (float idx) /. (float n) in
    let (x, y) = (r *. cos (s +. th0), r *. sin (s +. th0) +. dy) in
    if idx >= n then
      begin
	GlDraw.vertex3(x, y, z);
	GlDraw.ends()
      end
    else
      begin
	if idx = 0 then GlDraw.begins `polygon;
	GlDraw.vertex3(x, y, z);
	draw_circle z r th0 n (idx+1)
      end
  in

  let rec draw_cube (x0, y0, z0) lx ly lz =
    GlDraw.begins `quads;
    
    GlDraw.color (0.5,0.5,0.5);                
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0-.lz);
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0-.lz);
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0-.lz);
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0-.lz);  
    
    GlDraw.color (0.5,0.5,0.5);      
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0+.lz);
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0+.z);
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0-.z);
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0-.z);  
    
    GlDraw.color (0.5,0.5,0.5);
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0+.lz);
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0+.lz);
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0+.lz);
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0+.lz);
    
    GlDraw.color (0.5,0.5,0.5);
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0-.lz); 
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0-.lz); 
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0-.lz); 
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0-.lz); 
    
    GlDraw.color (0.5,0.5,0.5);                   
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0+.lz); 
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0-.lz); 
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0-.lz); 
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0+.lz); 
    
    GlDraw.color (0.5,0.5,0.5);         
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0-.lz); 
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0+.lz); 
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0+.lz); 
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0-.lz); 
    GlDraw.ends ()
  in
  
  let dy = cos !theta in
	
  GlClear.clear [`color; `depth];
  
  (* Draw cylinder *)
  GlMat.load_identity ();
  GlMat.translate3 (0., 0., -7.0);
  GlMat.rotate3 45.0 (0., 1., 0.);
  GlDraw.color (0.5, 0.5, 0.5);

  draw_cylinder (-5.) (-2.) 1. 1. !theta 30 0;
  draw_circle (-2.) 1. !theta 30 0; 
  GlDraw.color (0.55, 0.55, 0.55);
  draw_cylinder (-2.) (-1.0) 0.5 0.5 !theta 30 0;
  draw_circle (-1.0) 0.5 !theta 30 0;
  GlDraw.color (0.6, 0.6, 0.6);
  draw_cylinder (-1.0) (-0.5) 0.25 0.25 !theta 30 0;
  GlDraw.color (0.65, 0.65, 0.65);
  draw_circle (-0.5) 0.25 !theta 30 0; 
  
  GlDraw.color (0.7, 0.7, 0.7);
  draw_cylinder (-2.75) (-2.25) 1. 1. 0. 30 0;
  draw_cylinder (-2.75) (-2.25) 1.02 1.02 0. 30 0;
  draw_cylinder (-4.75) (-4.25) 1. 1. 0. 30 0;
  draw_cylinder (-4.75) (-4.25) 1.02 1.02 0. 30 0;
  
  draw_cube (1.,-.1., (0.5 *. (cos !theta) +. 0.5)) 1. 0.2 1.;
	
  Glut.swapBuffers ();
  theta := !theta +. omega


(* Handle window reshape events *)
let reshape_cb ~w ~h =
  let ratio = (float_of_int w) /. (float_of_int h) in
  GlDraw.viewport 0 0 w h;
  GlMat.mode `projection;
  GlMat.load_identity ();
  GluMat.perspective 45.0 ratio (0.1, 100.0);
  GlMat.mode `modelview;
  GlMat.load_identity ()

(* Handle keyboard events *)
let keyboard_cb ~key ~x ~y =
  match key with
  | 27 (* ESC *) -> exit 0
  | _ -> ()

(* Draw the scene whever idle *)
let idle_cb () =
  draw_gl_scene ()

let () =
  let width = 640 and height = 480 in
  ignore (Glut.init Sys.argv);
  Glut.initDisplayMode ~alpha:true ~depth:true ~double_buffer:true ();
  Glut.initWindowSize width height;
  ignore (Glut.createWindow "LablGL Machine Tool Simulation");
  Glut.displayFunc draw_gl_scene;
  Glut.keyboardFunc keyboard_cb;
  Glut.reshapeFunc reshape_cb;
  Glut.idleFunc(Some idle_cb);
  init_gl width height;
  Glut.mainLoop ()
